"""Vertex IDs of keypoints for various models."""
# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# You can only use this computer program if you have closed
# a license agreement with MPG or you get the right to use the computer
# program from someone who is authorized to grant you that right.
# Any use of the computer program without a valid license is prohibited and
# liable to prosecution.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# Contact: ps-license@tuebingen.mpg.de


vertex_ids = {
    'smal': {
        'lnose':	    1845,
        'reye':		    3044,
        'leye':		    1086,
        'rear_tip':		2508,
        'lear_tip':		543,
        'jaw':		    11,
        'tail':		    1836,
    },
    'hsmal': {
        'lnostril':		575,
        'reye':		    1191,
        'leye':		    378,
        'rear_tip':		1039,
        'lear_tip':		226,
        'jaw':		    611,
        'tail':		    752,
    },
    # NEED UPDATING. CURRENTLY SAME AS VAREN
    'hsmal+': {
        'nose':		    1738,
        'reye':		    3498,
        'leye':		    1596,
        'rear_tip':		2769,
        'lear_tip':		853,
        'jaw':		    1732,
        'tail':		    1910,
    },
    'varen': {
        'lfront_foot_tip': 1473,
        'rfront_foot_tip': 3390,
        'lfront_sesamoid': 1505,
        'rfront_sesamoid': 3422,
        'lback_foot_tip': 1410,
        'rback_foot_tip': 3327,
        'lback_calcaneous': 744,
        'rback_calcaneous': 2661,
        'back_left_foot_back': 1359,
        'back_right_foot_back': 3276,
        'front_left_foot_back': 1446,
        'front_right_foot_back': 3363,
        'lfront_knee': 663,
        'rfront_knee': 2580,
        'lear_tip': 852,
        'rear_tip': 2769,
        'tail_start': 1717,
        'tail_tip': 1915,
        'half_tail': 1899,
        'lear_base_back': 798,
        'rear_base_back': 2714,
        'lear_base_front_internal': 881,  # 884,
        'lear_base_front_external': 904,  # 1314,
        'rear_base_front_internal': 2798,  # 2801,
        'rear_base_front_external': 2821,  # 3231,
        'between_nostrils_base': 1739,
        'lnostril': 1154,
        'rnostril': 3071,  # 3071,
        'leye': 1250,  # 1583,
        'reye': 3498,
        'rcalcaneous': 2661,
        'lcalcaneous': 744,
        'lfront_coffin': 1473,
        'rfront_coffin': 3390,
        'lback_coffin': 1410,
        'rback_coffin': 3327,
        'end_sacral_vertebrae': 1717,
        'rcarpus_back': 2005,
        'lcarpus_back': 88,
        'tip_of_head': 1691,
        'rposterior': 2536,  # 2176,
        'lposterior': 619,  # 259,
        'middle_back': 1712

        }
    # 'varen': {
    #     'nose':		    1738,
    #     'reye':		    3498,
    #     'leye':		    1596,
    #     'rear':		    2769,
    #     'lear':		    853,
    #     'jaw':		    1732,
    #     'tail':		    1910,
    # }
}



"""Collection of utility functions and classes for the VARENA model."""
# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# You can only use this computer program if you have closed
# a license agreement with MPG or you get the right to use the computer
# program from someone who is authorized to grant you that right.
# Any use of the computer program without a valid license is prohibited and
# liable to prosecution.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# Contact: ps-license@tuebingen.mpg.de

from dataclasses import dataclass, fields
from typing import NewType, Optional, Union

import numpy as np
import torch

Tensor = NewType('Tensor', torch.Tensor)
Array = NewType('Array', np.ndarray)


@dataclass
class ModelOutput:
    """Dataclass to hand the varen Output.

    Attributes:
        vertices: The vertices of the SMPL model.
        joints: The joints of the SMPL model.
        full_pose: The full pose of the SMPL model.
        global_orient: The global orientation of the SMPL model.
        transl: The translation of the SMPL model.

    Note:
        The attributes are optional and can be None.
        The attributes are not guaranteed to be present in the output.

    """

    vertices: Optional[Tensor] = None
    joints: Optional[Tensor] = None
    full_pose: Optional[Tensor] = None
    global_orient: Optional[Tensor] = None
    transl: Optional[Tensor] = None

    def __getitem__(self, key):
        """Return the value of the attribute with the given key."""
        return getattr(self, key)

    def get(self, key, default=None):
        """Return the value of the attribute with the given key."""
        return getattr(self, key, default)

    def __iter__(self):
        """Return an iterator over the keys of the attributes."""
        return self.keys()

    def keys(self):
        """Return an iterator over the keys of the attributes."""
        keys = [t.name for t in fields(self)]
        return iter(keys)

    def values(self):
        """Return an iterator over the values of the attributes."""
        values = [getattr(self, t.name) for t in fields(self)]
        return iter(values)

    def items(self):
        """Return an iterator over the keys and values of the attributes."""
        data = [(t.name, getattr(self, t.name)) for t in fields(self)]
        return iter(data)


@dataclass
class MuscleDeformer:
    """Dataclass to handle the muscle deformer output.

    Attributes:
        betas_muscle: The muscle deformation parameters.
        Bm: The muscle deformation matrix.
        muscle_idxs: The indices of the muscles.

    """

    # betas_muscle, Bm, muscle_idxs. Not Optional:
    betas_muscle: Tensor
    Bm: Tensor
    muscle_idxs: Tensor


@dataclass
class SMALOutput(ModelOutput):
    """Dataclass to handle the SMAL output.

    Attributes:
        betas: The shape parameters of the SMAL model.
        body_pose: The pose parameters of the SMAL model.

    """

    betas: Optional[Tensor] = None
    body_pose: Optional[Tensor] = None


@dataclass
class VARENOutput(ModelOutput):
    """Dataclass to handle the VAREN output.

    Attributes:
        mdv: Muscle Deformation 3D Offsets.
        surface_keypoints: The surface keypoints of the VAREN model.
        body_pose: The pose parameters of the VAREN model.
        body_betas: The shape parameters of the VAREN model.
        muscle_betas: The muscle deformation parameters.
        muscle_activations: The muscle activations of the VAREN model.

    """

    mdv: Optional[Tensor] = None
    surface_keypoints: Optional[Tensor] = None,
    body_pose: Optional[Tensor] = None
    body_betas: Optional[Tensor] = None
    muscle_betas: Optional[Tensor] = None
    muscle_activations: Optional[Tensor] = None


def find_joint_kin_chain(joint_id, kinematic_tree):
    """Find the kinematic chain of a joint.

    Args:
        joint_id: The id of the joint.
        kinematic_tree: The kinematic tree of the model.

    Returns:
        kin_chain: The kinematic chain of the joint.

    """
    kin_chain = []
    curr_idx = joint_id
    while curr_idx != -1:
        kin_chain.append(curr_idx)
        curr_idx = kinematic_tree[curr_idx]
    return kin_chain


def to_tensor(
        array: Union[Array, Tensor], dtype=torch.float32
) -> Tensor:
    """Convert a numpy array to a torch tensor."""
    if torch.is_tensor(array):
        return array
    return torch.tensor(array, dtype=dtype)


class Struct:
    """A simple class to create a struct-like object.

    This class allows you to create an object with attributes
    that can be accessed like a struct.

    Example:
        obj = Struct(a=1, b=2)
        print(obj.a)  # 1
        print(obj.b)  # 2

    """

    def __init__(self, **kwargs):
        """Initialise the struct with the given keyword arguments."""
        for key, val in kwargs.items():
            setattr(self, key, val)


def to_np(array, dtype=np.float32):
    """Convert a tensor or space array to a numpy array."""
    if 'scipy.sparse' in str(type(array)):
        array = array.todense()
    return np.array(array, dtype=dtype)


def rot_mat_to_euler(rot_mats):
    """Convert rotation matrices to euler angles.

    Args:
        rot_mats: Rotation matrices of shape (..., 3, 3).

    Returns (torch.Tensor):
        euler angles of shape (..., 3).

    """
    # Calculates rotation matrix to euler angles
    # Careful for extreme cases of eular angles like [0.0, pi, 0.0]

    sy = torch.sqrt(rot_mats[:, 0, 0] * rot_mats[:, 0, 0] +
                    rot_mats[:, 1, 0] * rot_mats[:, 1, 0])
    return torch.atan2(-rot_mats[:, 2, 0], sy)


def axis_angle_to_quaternion(axis_angle: torch.Tensor) -> torch.Tensor:
    """Convert rotations given as axis/angle to quaternions.

    Args:
        axis_angle: Rotations given as a vector in axis angle form,
            as a tensor of shape (..., 3), where the magnitude is
            the angle turned anticlockwise in radians around the
            vector's direction.

    Returns:
        quaternions with real part first, as tensor of shape (..., 4).

    Taken from pytorch3d.transforms

    """
    angles = torch.norm(axis_angle, p=2, dim=-1, keepdim=True)
    half_angles = angles * 0.5
    eps = 1e-6
    small_angles = angles.abs() < eps
    sin_half_angles_over_angles = torch.empty_like(angles)
    sin_half_angles_over_angles[~small_angles] = (
        torch.sin(half_angles[~small_angles]) / angles[~small_angles]
    )
    # for x small, sin(x/2) is about x/2 - (x/2)^3/6
    # so sin(x/2)/x is about 1/2 - (x*x)/48
    sin_half_angles_over_angles[small_angles] = (
        0.5 - (angles[small_angles] * angles[small_angles]) / 48
    )
    quaternions = torch.cat(
        [
        torch.cos(half_angles),
        axis_angle * sin_half_angles_over_angles
        ],
        dim=-1
    )
    return quaternions



import pytest
import torch

from varen import VAREN
from varen.utils import VARENOutput


def test_forward_pass_full(varen_model):
    output = varen_model()
    assert isinstance(output, VARENOutput), "Output should be a VARENOutput instance."

    expected_keys = [
        "vertices",
        "joints",
        "mdv",
        "muscle_betas",
        "muscle_activations",
        "body_betas",
        "body_pose",
        "surface_keypoints",
        "full_pose",
        "global_orient"
    ]

    # Collect missing keys
    missing_keys = [key for key in expected_keys if key not in output]

    # Collect keys that should be None but aren't
    should_be_not_none_keys = ["mdv", "muscle_betas", "muscle_activations"]
    none_keys = [key for key in should_be_not_none_keys if output.get(key) is None]

    # Raise errors if any problems found
    errors = []
    if missing_keys:
        errors.append(f"Missing keys: {', '.join(missing_keys)}")
    if none_keys:
        errors.append(f"Expected None values for: {', '.join(none_keys)}")

    if errors:
        raise AssertionError(" | ".join(errors))

def test_forward_pass_no_muscles(varen_model_no_muscles):
    output = varen_model_no_muscles()
    assert isinstance(output, VARENOutput), "Output should be a VARENOutput instance."

    expected_keys = [
        "vertices",
        "joints",
        "mdv",
        "muscle_betas",
        "muscle_activations",
        "body_betas",
        "body_pose",
        "surface_keypoints",
        "full_pose",
        "global_orient"
    ]

    # Collect missing keys
    missing_keys = [key for key in expected_keys if key not in output]

    # Collect keys that should be None but aren't
    should_be_none_keys = ["mdv", "muscle_betas", "muscle_activations"]
    non_none_values = {
        key: output.get(key) for key in should_be_none_keys if output.get(key) is not None
    }

    # Raise errors if any problems found
    errors = []
    if missing_keys:
        errors.append(f"Missing keys: {', '.join(missing_keys)}")
    if non_none_values:
        for key, val in non_none_values.items():
            errors.append(f"Expected '{key}' to be None but got: {val}")

    if errors:
        raise AssertionError(" | ".join(errors))


def test_vertex_output(varen_model):
    output = varen_model()
    assert isinstance(output, VARENOutput), "Output should be a VARENOutput instance."
    vertices = output.vertices
    template = varen_model.v_template

    assert vertices is not None, "Vertices should not be None."
    assert isinstance(vertices, torch.Tensor), f"Vertices should be a torch.Tensor, got {type(vertices)}."
    assert vertices.dim() == 3, f"Vertices should have 3 dimensions. Got: {vertices.dim()}."
    assert vertices.shape[1:] == template.shape, f"Vertices shape should match template shape {template.shape}, got: {vertices.shape[1:]}."
    assert torch.isfinite(vertices).all(), "Vertices should not contain NaN or Inf values."

def test_joints_output(varen_model):
    output = varen_model()
    assert isinstance(output, VARENOutput), "Output should be a VARENOutput instance."
    joints = output.joints

    assert joints is not None, "Joints should not be None."
    assert isinstance(joints, torch.Tensor), f"Joints should be a torch.Tensor, got {type(joints)}."
    assert joints.dim() == 3, f"Joints should have 3 dimensions. Got: {joints.dim()}."
    # NOTE: We include + 1 for the global joint at location 0. In the model
    # files, we have a double up on joints 0 and 1. Both 'spine' and 'pelvis'
    # are located at 0,0,0. Overall we have 37 UNIQUE joints, but 38 joints
    # overall... The body_pose vector does not rotate around the GO, so it is
    # 37 long. Once we add the global joint, we have 38 'joints'.

    # NOTE 2: To change this, we would either want to move one of the joints,
    # so that there is no double up, OR we remove one of them and adjust the
    # sizing of everything.
    assert joints.shape[1] == varen_model.NUM_JOINTS + 1, f"Joints shape should match template joints shape {varen_model.NUM_JOINTS + 1}, got: {joints.shape}."
    assert torch.isfinite(joints).all(), "Joints should not contain NaN or Inf values."


import pytest
from varen import VAREN
from pathlib import Path


def test_loading_without_muscles(varen_model_no_muscles):
    varen = varen_model_no_muscles
    assert isinstance(varen, VAREN), "Model is not an instance of VAREN"


def test_loading(varen_model):
    varen = varen_model
    assert isinstance(varen, VAREN), "Model is not an instance of VAREN"

def test_varen_model_loading_bad_model_dir(
        varen_model_dir, varen_model_file_name, ckpt_file
    ):
    """Test to see if model fails to load when bad dir is passed."""
    bad_model_dir = "/non/existent/path"
    bad_model_file_name = "missing_model"
    bad_ckpt_file = "/checkpoint_not_real.pth"

    varen_pkl = Path(varen_model_file_name)
    path, ext = varen_pkl.stem, varen_pkl.suffix

    with pytest.raises(FileNotFoundError):
        VAREN(
            model_path=bad_model_dir,
            model_file_name=path,
            use_muscle_deformations=False,
            ext=ext
        )

def test_varen_model_loading_bad_model_filename(
        varen_model_dir, varen_model_file_name, ckpt_file
    ):
    """Test to see if model fails to load when bad filename is passed."""
    bad_model_dir = "/non/existent/path"
    bad_model_file_name = "missing_model_45812187"
    bad_ckpt_file = "/checkpoint_not_real.pth"

    varen_pkl = Path(varen_model_file_name)
    path, ext = varen_pkl.stem, varen_pkl.suffix
    with pytest.raises(FileNotFoundError):
        VAREN(
            model_path=varen_model_dir,
            model_file_name=bad_model_file_name,
            use_muscle_deformations=False,
        )

def test_varen_model_loading_bad_model_ext(
        varen_model_dir, varen_model_file_name, ckpt_file
    ):
    """Test to see if model fails to load when bad filename is passed."""

    varen_pkl = Path(varen_model_file_name)
    path, ext = varen_pkl.stem, varen_pkl.suffix

    with pytest.raises(FileNotFoundError):
        VAREN(
            model_path=varen_model_dir,
            model_file_name=path,
            use_muscle_deformations=False,
            ext="notarealext"
        )

def test_varen_model_loading_bad_ckpt_filename(
        varen_model_dir, varen_model_file_name, ckpt_file
    ):
    """Test to see if model fails to load when bad filename is passed."""

    varen_pkl = Path(varen_model_file_name)
    path, ext = varen_pkl.stem, varen_pkl.suffix

    with pytest.raises(FileNotFoundError):
        VAREN(
            model_path=varen_model_dir,
            model_file_name=path,
            use_muscle_deformations=True,
            ext=ext,
            ckpt_file="/non/existent/path/to/ckpt.pth"

        )


"""Collection of classes for horse models."""
import os
import os.path as osp
import pickle
import warnings
from collections import namedtuple
from typing import Optional, Union

import numpy as np
import torch
from torch import nn

from .lbs import blend_shapes, lbs
from .utils import (
    Array,
    MuscleDeformer,
    SMALOutput,
    Struct,
    Tensor,
    VARENOutput,
    axis_angle_to_quaternion,
    to_np,
    to_tensor,
)
from .vertex_ids import vertex_ids as VERTEX_IDS
from .vertex_joint_selector import VertexJointSelector

TensorOutput = namedtuple('TensorOutput',
                          [
                            'vertices', 'joints', 'betas', 'expression',
                            'global_orient', 'body_pose', 'left_hand_pose',
                            'right_hand_pose', 'jaw_pose', 'transl',
                            'full_pose'
                            ]
                        )


class SMAL(nn.Module):
    """Original quadrupedal model.

    Based on implementation from: SMAL paper. Which in turn is based on
    the SMPL implementation.
    """

    NUM_JOINTS = 32
    SHAPE_SPACE_DIM = 10

    def __init__(
        self, model_path: str,
        data_struct: Optional[Struct] = None,
        create_betas: bool = True,
        betas: Optional[Tensor] = None,
        num_betas: int = 10,
        create_global_orient: bool = True,
        global_orient: Optional[Tensor] = None,
        create_body_pose: bool = True,
        body_pose: Optional[Tensor] = None,
        create_transl: bool = True,
        transl: Optional[Tensor] = None,
        dtype=torch.float32,
        batch_size: int = 1,
        joint_mapper=None,
        vertex_ids: dict[str, int] = None,
        v_template: Optional[Union[Tensor, Array]] = None,
        device='cpu',
        **kwargs
    ) -> None:
        """SMPL model constructor.

        Parameters
        ----------
            model_path: str
                The path to the folder or to the file where the model
                parameters are stored
            data_struct: Strct
                A struct object. If given, then the parameters of the model are
                read from the object. Otherwise, the model tries to read the
                parameters from the given `model_path`. (default = None)
            create_global_orient: bool, optional
                Flag for creating a member variable for the global orientation
                of the body. (default = True)
            global_orient: torch.tensor, optional, Bx3
                The default value for the global orientation variable.
                (default = None)
            create_body_pose: bool, optional
                Flag for creating a member variable for the pose of the body.
                (default = True)
            body_pose: torch.tensor, optional, Bx(Body Joints * 3)
                The default value for the body pose variable.
                (default = None)
            num_betas: int, optional
                Number of shape components to use
                (default = 10).
            create_betas: bool, optional
                Flag for creating a member variable for the shape space
                (default = True).
            betas: torch.tensor, optional, Bx10
                The default value for the shape member variable.
                (default = None)
            create_transl: bool, optional
                Flag for creating a member variable for the translation
                of the body. (default = True)
            transl: torch.tensor, optional, Bx3
                The default value for the transl variable.
                (default = None)
            dtype: torch.dtype, optional
                The data type for the created variables
            batch_size: int, optional
                The batch size used for creating the member variables
            joint_mapper: object, optional
                An object that re-maps the joints. Useful if one wants to
                re-order the model joints to some other convention
                (e.g. MSCOCO) (default = None)
            gender: str, optional
                Which gender to load
            vertex_ids: dict, optional
                A dictionary containing the indices of the extra vertices that
                will be selected

        """
        # NOTE: No PCA on the pose space

        if data_struct is None:
            if osp.isdir(model_path):
                model_fn = 'SMAL.{ext}'.format(ext='pkl')
                smpl_path = os.path.join(model_path, model_fn)
            else:
                smpl_path = model_path
            if not osp.exists(smpl_path):
                raise FileNotFoundError(f'Path {smpl_path} does not exist!')

            with open(smpl_path, 'rb') as smpl_file:
                data_struct = Struct(**pickle.load(smpl_file,
                                                   encoding='latin1'))

        super().__init__()
        self.batch_size = batch_size
        shapedirs = data_struct.shapedirs
        if (shapedirs.shape[-1] < self.SHAPE_SPACE_DIM):
            print(f'WARNING: You are using a {self.name()} model, with only'
                  f' {shapedirs.shape[-1]} shape coefficients.\n'
                  f'num_betas={num_betas}, shapedirs.shape={shapedirs.shape}, '
                  f'self.SHAPE_SPACE_DIM={self.SHAPE_SPACE_DIM}')
            num_betas = min(num_betas, shapedirs.shape[-1])
        else:
            num_betas = min(num_betas, self.SHAPE_SPACE_DIM)

        self._num_betas = num_betas
        shapedirs = shapedirs[:, :, :num_betas]

        # The shape components
        self.register_buffer(
            'shapedirs',
            to_tensor(to_np(shapedirs), dtype=dtype))

        if vertex_ids is None:
            vertex_ids = VERTEX_IDS['smal']

        self.dtype = dtype

        self.joint_mapper = joint_mapper

        # TO DO: Remove extra args.
        self.vertex_joint_selector = VertexJointSelector(
            vertex_ids=vertex_ids,
            use_feet_keypoints=False,
            use_hands=False,
            **kwargs
            )

        self.faces = data_struct.f
        self.register_buffer('faces_tensor',
                             to_tensor(to_np(self.faces, dtype=np.int64),
                                       dtype=torch.long))

        if create_betas:
            if betas is None:
                default_betas = torch.zeros(
                    [batch_size, self.num_betas], dtype=dtype, device=device)
            elif torch.is_tensor(betas):
                default_betas = betas.clone().detach()
            else:
                default_betas = torch.tensor(betas, dtype=dtype, device=device)

            self.register_parameter(
                'betas', nn.Parameter(default_betas, requires_grad=True))

        # The tensor that contains the global rotation of the model
        # It is separated from the pose of the joints in case we wish to
        # optimize only over one of them
        if create_global_orient:
            if global_orient is None:
                default_global_orient = torch.zeros(
                    [batch_size, 3], dtype=dtype, device=device)
            elif torch.is_tensor(global_orient):
                default_global_orient = global_orient.clone().detach()
            else:
                default_global_orient = torch.tensor(
                    global_orient, dtype=dtype, device=device)

            global_orient = nn.Parameter(default_global_orient,
                                         requires_grad=True)
            self.register_parameter('global_orient', global_orient)

        if create_body_pose:
            if body_pose is None:
                default_body_pose = torch.zeros(
                    [batch_size, self.NUM_JOINTS * 3],
                    dtype=dtype,
                    device=device
                    )
            elif torch.is_tensor(body_pose):
                default_body_pose = body_pose.clone().detach()
            else:
                default_body_pose = torch.tensor(body_pose,
                                                 dtype=dtype, device=device)
            self.register_parameter(
                'body_pose',
                nn.Parameter(default_body_pose, requires_grad=True))

        if create_transl:
            if transl is None:
                default_transl = torch.zeros(
                    [batch_size, 3],
                    dtype=dtype,
                    requires_grad=True,
                    device=device
                    )
            else:
                default_transl = torch.tensor(
                    transl,
                    dtype=dtype,
                    device=device
                    )
            self.register_parameter(
                'transl', nn.Parameter(default_transl, requires_grad=True))

        if v_template is None:
            v_template = data_struct.v_template

        if not torch.is_tensor(v_template):
            v_template = to_tensor(to_np(v_template), dtype=dtype)

        # The vertices of the template model
        self.register_buffer('v_template', v_template)

        j_regressor = to_tensor(to_np(
            data_struct.J_regressor), dtype=dtype)
        self.register_buffer('J_regressor', j_regressor)

        num_pose_basis = data_struct.posedirs.shape[-1]
        posedirs = np.reshape(data_struct.posedirs, [-1, num_pose_basis]).T
        self.register_buffer('posedirs',
                             to_tensor(to_np(posedirs), dtype=dtype))

        # indices of parents for each joints
        parents = to_tensor(to_np(data_struct.kintree_table[0])).long()
        parents[0] = -1
        self.register_buffer('parents', parents)

        lbs_weights = to_tensor(to_np(data_struct.weights), dtype=dtype)
        self.register_buffer('lbs_weights', lbs_weights)

    @property
    def num_betas(self):
        """Return the number of shape parameters."""
        return self._num_betas

    def create_mean_pose(self, data_struct) -> Tensor:
        """Create the mean pose of the model."""
        pass

    @staticmethod
    def name() -> str:
        """Return the name of the model."""
        return 'SMAL'

    @torch.no_grad()
    def reset_params(self, **params_dict) -> None:
        """Reset all parameters of the model to 0."""
        for param_name, param in self.named_parameters():
            if param_name in params_dict:
                param[:] = torch.tensor(params_dict[param_name])
            else:
                param.fill_(0)

    def get_num_verts(self) -> int:
        """Return the number of vertices in the model."""
        return self.v_template.shape[0]

    def get_num_faces(self) -> int:
        """Return the number of faces of the model."""
        return self.faces.shape[0]

    def get_num_joints(self) -> int:
        """Return the number of joints of the model."""
        return self.NUM_JOINTS

    def extra_repr(self) -> str:
        """Return the extended representation of the model."""
        msg = [
            f'Number of joints: {self.J_regressor.shape[0]}',
            f'Betas: {self.num_betas}',
        ]
        return '\n'.join(msg)

    def forward_shape(
        self,
        betas: Optional[Tensor] = None,
    ) -> SMALOutput:
        """Add shape contribution to the model template."""
        betas = betas if betas is not None else self.betas
        v_shaped = self.v_template + blend_shapes(betas, self.shapedirs)
        return SMALOutput(vertices=v_shaped, betas=betas)

    def forward(  # noqa: PLR0913, PLR0917
        self,
        betas: Optional[Tensor] = None,
        body_pose: Optional[Tensor] = None,
        global_orient: Optional[Tensor] = None,
        transl: Optional[Tensor] = None,
        return_verts=True,
        return_full_pose: bool = False,
        pose2rot: bool = True,
        **kwargs  # noqa: ARG002
    ) -> SMALOutput:
        """Forward pass for the SMAL model.

        Args:
            betas (Optional[Tensor], optional):
                Shape coefficients for the body model (default is None).
            body_pose (Optional[Tensor], optional):
                Body pose parameters (default is None).
            global_orient (Optional[Tensor], optional):
                Global orientation parameters (default is None).
            transl (Optional[Tensor], optional):
                Translation parameters (default is None).
            return_verts (bool, optional):
                Whether to return the vertices of the model (default is True).
            return_full_pose (bool, optional):
                Whether to return the full pose including global orientation
                (default is False).
            pose2rot (bool, optional):
                Whether to convert pose parameters to rotation matrices
                (default is True).
            **kwargs:
                Additional arguments passed to the method.

        Returns:
            SMALOutput: The output of the forward pass, including the computed
            body shape, pose, vertices, etc.

        """
        # If no shape and pose parameters are passed along, then use the
        # ones from the module
        global_orient = (global_orient if global_orient is not None else
                         self.global_orient)
        body_pose = body_pose if body_pose is not None else self.body_pose
        betas = betas if betas is not None else self.betas

        apply_trans = transl is not None or hasattr(self, 'transl')
        if transl is None and hasattr(self, 'transl'):
            transl = self.transl

        full_pose = torch.cat([global_orient, body_pose], dim=1)

        batch_size = max(betas.shape[0], global_orient.shape[0],
                         body_pose.shape[0])

        if betas.shape[0] != batch_size:
            num_repeats = int(batch_size / betas.shape[0])
            betas = betas.expand(num_repeats, -1)

        vertices, joints = lbs(betas, full_pose, self.v_template,
                               self.shapedirs, self.posedirs,
                               self.J_regressor, self.parents,
                               self.lbs_weights, pose2rot=pose2rot)

        # Add extra points to the joints (eg keypoints)?
        joints = self.vertex_joint_selector(vertices, joints)
        # Map the joints to the current dataset
        if self.joint_mapper is not None:
            joints = self.joint_mapper(joints)

        if apply_trans:
            joints += transl.unsqueeze(dim=1)
            vertices += transl.unsqueeze(dim=1)

        output = SMALOutput(vertices=vertices if return_verts else None,
                            global_orient=global_orient,
                            body_pose=body_pose,
                            joints=joints,
                            betas=betas,
                            full_pose=full_pose if return_full_pose else None)

        return output


class HSMAL(SMAL):
    """HSMAL (Horse-SMAL) model. Same as the SMAL model but for horses."""

    NUM_JOINTS = 35
    """SMAL with more vertices and faces"""
    # Just need to alter the vertex IDs. The rest is the same
    def __init__(  # noqa: PLR0913, PLR0917
            self,
            model_path: str,
            data_struct: Optional[Struct] = None,
            dtype=torch.float32,
            batch_size: int = 1,
            joint_mapper=None,
            vertex_ids: dict[str, int] = None,
            v_template: Optional[Union[Tensor, Array]] = None,
            ext: str = 'pkl',
            high_res: bool = True,
            **kwargs) -> None:
        """HSMAL Constructor. Initialise the HSMAL model.

        Args:
            model_path (str):
                Path to the model file directory.
            data_struct (Optional[Struct], optional):
                The data structure containing model information
                (default is None).
            dtype (torch.dtype, optional):
                The data type of tensors (default is torch.float32).
            batch_size (int, optional):
                The batch size for the model (default is 1).
            joint_mapper (Optional[callable], optional):
                A function to map joints (default is None).
            vertex_ids (Optional[Dict[str, int]], optional):
                A dictionary of vertex IDs for different parts of the horse's
                body (default is None).
            v_template (Optional[Union[Tensor, Array]], optional):
                Template vertex data for the model (default is None).
            ext (str, optional):
                File extension for the model data (default is 'pkl').
            high_res (bool, optional):
                Whether to use high-resolution model data (default is True).
            **kwargs:
                Additional arguments passed to the parent class initialization.

        """
        self.high_res = high_res
        if data_struct is None:
            if osp.isdir(model_path):
                model_fn = 'HSMAL{}.{ext}'.format(
                    "+" if self.high_res else "", ext='pkl'
                    )
                hsmal_path = os.path.join(model_path, model_fn)
            else:
                hsmal_path = model_path
            assert osp.exists(hsmal_path), f'Path {hsmal_path} does not exist!'

            with open(hsmal_path, 'rb') as smpl_file:
                data_struct = Struct(**pickle.load(smpl_file,
                                                   encoding='latin1'))

        if vertex_ids is None:
            if self.high_res:
                vertex_ids = VERTEX_IDS['hsmal+']
            else:
                vertex_ids = VERTEX_IDS['hsmal']

        super().__init__(
            model_path=model_path,
            data_struct=data_struct,
            batch_size=batch_size,
            v_template=v_template,
            joint_mapper=joint_mapper,
            dtype=dtype,
            vertex_ids=vertex_ids,
            ext=ext,
            **kwargs)

        self.vertex_joint_selector = VertexJointSelector(
            vertex_ids=vertex_ids,
            use_hands=False,
            use_feet_keypoints=False,
            **kwargs)


class VAREN(HSMAL):
    """VAREN class that extends the HSMAL class.

    This model is designed for equine (horse) body shape and pose estimation.
    It utilizes a similar approach to SMPL, but is specifically tailored
    for horses. The model includes the ability to simulate muscle deformations
    using a neural  network, making it suitable for accurate and realistic
    representations of equine structures and movements.

    Args:
        model_path (str):
            Path to the model directory or file.
        data_struct (Optional[Struct], optional):
            The data structure containing model information (default is None).
        num_betas (int, optional):
            The number of shape coefficients to use for the model
            (default is 39).
        use_muscle_deformations (bool, optional):
            Whether to include muscle deformations (default is True).
        shape_betas_for_muscles (int, optional):
            Number of shape betas for muscle modeling (default is 2).
        muscle_betas_size (int, optional):
            The size of muscle betas (default is 1).
        dtype (torch.dtype, optional):
            The data type of tensors (default is torch.float32).
        batch_size (int, optional):
            The batch size for the model (default is 1).
        joint_mapper (Optional[callable], optional):
            A function to map joints (default is None).
        vertex_ids (Optional[Dict[str, int]], optional):
            A dictionary of vertex IDs for different parts of the horse's body
            (default is None).
        v_template (Optional[Union[Tensor, Array]], optional):
            Template vertex data for the model (default is None).
        ext (str, optional):
            File extension for the model data (default is 'pkl').
        model_file_name (Optional[str], optional):
            The file name for the model (default is None).
        ckpt_file (Optional[str], optional):
            Path to a checkpoint file for loading the model (default is None).
        **kwargs:
            Additional arguments passed to the parent class initialization.

    """

    NUM_JOINTS = 37  # Results in 38 joints including 0
    SHAPE_SPACE_DIM = 39  # The dimensionality of the shape space for the model

    def __init__(self, model_path: str,  # noqa: PLR0913, PLR0917
                 data_struct: Optional[Struct] = None,
                 num_betas: int = 39,
                 use_muscle_deformations: bool = True,
                 shape_betas_for_muscles: int = 2,
                 muscle_betas_size: int = 1,
                 dtype=torch.float32,
                 batch_size: int = 1,
                 joint_mapper=None,
                 vertex_ids: dict[str, int] = None,
                 v_template: Optional[Union[Tensor, Array]] = None,
                 ext: str = 'pkl',
                 model_file_name: Optional[str] = None,
                 ckpt_file: Optional[str] = 'varen.pth',
                 **kwargs) -> None:
        """Initialize the VAREN model.

        Args:
            model_path (str):
                The path to the directory containing the model data or the
                model file.
            data_struct (Optional[Struct], optional):
                Data structure containing model details (default is None).
            num_betas (int, optional):
                Number of shape coefficients for the model (default is 39).
            use_muscle_deformations (bool, optional):
                Whether to use muscle deformations (default is True).
            shape_betas_for_muscles (int, optional):
                Number of shape betas for muscle modeling (default is 2).
            muscle_betas_size (int, optional):
                The size of muscle betas (default is 1).
            dtype (torch.dtype, optional):
                Data type for the model tensors (default is torch.float32).
            batch_size (int, optional):
                Batch size (default is 1).
            joint_mapper (Optional[callable], optional):
                A mapping function for joints (default is None).
            vertex_ids (Optional[Dict[str, int]], optional):
                Dictionary mapping body parts to vertex IDs (default is None).
            v_template (Optional[Union[Tensor, Array]], optional):
                Template vertices (default is None).
            ext (str, optional):
                File extension for the model (default is 'pkl').
            model_file_name (Optional[str], optional):
                The specific model filename (default is None).
            ckpt_file (Optional[str], optional):
                Path to the checkpoint file for model loading
                (default is None).
            **kwargs:
                Additional arguments passed to the parent class HSMAL
                initialization.

        """
        self.use_muscle_deformations = use_muscle_deformations
        self.shape_betas_for_muscles = shape_betas_for_muscles
        self.muscle_betas_size = muscle_betas_size

        if data_struct is None:
            if osp.isdir(model_path):

                model_fn = '{}.{ext}'.format(
                    "VAREN" if model_file_name is None
                    else model_file_name, ext=ext.replace('.', '')
                    )
                varen_path = os.path.join(model_path, model_fn)
            else:
                varen_path = model_path
            if not osp.exists(varen_path):
                raise FileNotFoundError(f'Path {varen_path} does not exist!')

            with open(varen_path, 'rb') as file:
                data_struct = Struct(**pickle.load(file,
                                                   encoding='latin1'))

        if vertex_ids is None:
            vertex_ids = VERTEX_IDS['varen']

        super().__init__(
            model_path=model_path,
            data_struct=data_struct,
            batch_size=batch_size,
            v_template=v_template,
            joint_mapper=joint_mapper,
            dtype=dtype,
            vertex_ids=vertex_ids,
            ext=ext,
            high_res=True,
            num_betas=num_betas,
            **kwargs)

        self.batch_size = batch_size
        shapedirs = data_struct.shapedirs
        if (shapedirs.shape[-1] < self.SHAPE_SPACE_DIM):
            print(f'WARNING: You are using a {self.name()} model, with only'
                  f' {shapedirs.shape[-1]} shape coefficients.\n'
                  f'num_betas={num_betas}, shapedirs.shape={shapedirs.shape}, '
                  f'self.SHAPE_SPACE_DIM={self.SHAPE_SPACE_DIM}')
            num_betas = min(num_betas, shapedirs.shape[-1])
        else:
            num_betas = min(num_betas, self.SHAPE_SPACE_DIM)

        self._num_betas = num_betas
        shapedirs = shapedirs[:, :, :num_betas]
        # The shape components
        self.register_buffer(
            'shapedirs',
            to_tensor(to_np(shapedirs), dtype=dtype))

        # Add additional information about the part segmentation
        if hasattr(data_struct, 'parts'):
            self.parts = data_struct.parts
            self.partSet = range(len(self.parts))

        # Likely depricated for the below
        if hasattr(data_struct, 'part2vertices'):
            self.part2vertices = data_struct.part2vertices

        if hasattr(data_struct, 'colors_names'):
            self.colors_names = data_struct.colors_names

        if hasattr(data_struct, 'seg'):
            self.seg = data_struct.seg

        if hasattr(data_struct, 'muscle_labels'):
            self.muscle_labels = data_struct.muscle_labels

        self.vertex_joint_selector.extra_joints_idxs = to_tensor(
            list(VERTEX_IDS['varen'].values()), dtype=torch.long)

        if self.use_muscle_deformations:
            # Neural muscle deformer added to self
            self.create_neural_muscle_deformer()

            # If path exists, load checkpoint
            if ckpt_file is not None:
                ckpt_path = osp.join(model_path, ckpt_file)
            else:
                ckpt_path = None

            if ckpt_path is not None:
                print("Loading VAREN Muscle Model from: ", ckpt_path)
                chkpt = torch.load(ckpt_path, weights_only=True)
                # Load Bm weights (Muscle Betas) -> Deforms vertices based on
                # betas
                self.Bm.load_state_dict(chkpt['Bm'])
                # Load betas muscle predictor weights
                self.betas_muscle_predictor.load_state_dict(
                    chkpt['betas_muscle_predictor']
                    )

    def create_neural_muscle_deformer(self) -> None:
        """Create and initialize the neural muscle deformation model.

        Uses muscle labels and associations. This method defines the
        muscle deformations based on the muscle labels and shape betas for
        muscles.

        """
        muscle_associations = self.define_muscle_deformations_variables()

        # Predict Muscle Betas based on pose and shape
        self.betas_muscle_predictor = MuscleBetaPredictor(
            muscle_associations=muscle_associations,
            shape_beta_for_muscles=self.shape_betas_for_muscles
        )

        # Predict offsets based on muscle betas
        self.Bm = self.create_Bm()

    def forward(self,
                betas: Optional[Tensor] = None,
                body_pose: Optional[Tensor] = None,
                global_orient: Optional[Tensor] = None,
                transl: Optional[Tensor] = None,
                return_verts: bool = True,
                return_full_pose: bool = False,
                pose2rot: bool = True,
                **kwargs) -> VARENOutput:  # noqa: ARG002
        """Deform the template using pose and shape.

        Args:
            betas (Optional[Tensor], optional):
                Shape coefficients for the body model (default is None).
            body_pose (Optional[Tensor], optional):
                Body pose parameters (default is None).
            global_orient (Optional[Tensor], optional):
                Global orientation parameters (default is None).
            transl (Optional[Tensor], optional):
                Translation parameters (default is None).
            return_verts (bool, optional):
                Whether to return the vertices of the model (default is True).
            return_full_pose (bool, optional):
                Whether to return the full pose including global orientation
                (default is False).
            pose2rot (bool, optional):
                Whether to convert pose parameters to rotation matrices
                (default is True).
            **kwargs: Additional arguments passed to the method.

        Returns:
            VARENOutput (Named Tuple):
            The output of the forward pass, including the computed
            body shape, pose, vertices, muscle activations, etc.

        """
        # Warn if some variant of 'pose' is passed, that isn't 'body_pose'
        if body_pose is None:
            for key in kwargs:
                if 'pose' in key.lower() and key != 'body_pose':
                    warnings.warn(
                        f"Unexpected key '{key}' in kwargs. "
                        "Did you mean 'body_pose'?",
                        UserWarning, stacklevel=1
                    )

        global_orient = self.global_orient if global_orient \
            is None else global_orient
        body_pose = self.body_pose if body_pose is None else body_pose
        betas = self.betas if betas is None else betas
        transl = self.transl if transl is None else transl

        full_pose = torch.cat([global_orient, body_pose], dim=1)

        # Muscle Predictor forward pass
        muscle_deformer = self.compute_muscle_deformations(full_pose, betas) \
            if self.use_muscle_deformations else None

        vertices, joints, mdv = lbs(
            betas,
            full_pose,
            self.v_template,
            self.shapedirs,
            self.posedirs,
            self.J_regressor,
            self.parents,
            self.lbs_weights,
            pose2rot=pose2rot,
            muscle_deformer=muscle_deformer
            )

        joints = self.vertex_joint_selector(vertices, joints)
        if self.joint_mapper is not None:
            joints = self.joint_mapper(joints)

        if transl is not None:
            joints += transl.unsqueeze(dim=1)
            vertices += transl.unsqueeze(dim=1)

        output = VARENOutput(vertices=vertices if return_verts else None,
                        global_orient=global_orient,
                        body_pose=body_pose,
                        joints=joints[:, :self.NUM_JOINTS+1],
                        surface_keypoints=joints[:, self.NUM_JOINTS+1:],
                        body_betas=betas,
                        muscle_betas=(
                            muscle_deformer.betas_muscle
                            if self.use_muscle_deformations else None
                        ),
                        full_pose=full_pose if return_full_pose else None,
                        muscle_activations=(
                            self.A
                            if self.use_muscle_deformations else None
                        ),
                        mdv=mdv if self.use_muscle_deformations else None)
        return output

    def compute_muscle_deformations(self, full_pose, betas):
        """Compute muscle deformations if enabled."""
        muscle_betas, self.A = self.betas_muscle_predictor.forward(
            full_pose,
            betas
        )
        # Muscle Deformer is a just a Dataclass containing the outputs of
        # MuscleBetasPredictor
        return MuscleDeformer(muscle_betas, self.Bm, self.muscle_vertex_map)

    def define_muscle_deformations_variables(self) -> torch.Tensor:
        """Define muscle deformation variables.

        Includes muscle associations and muscle vertex indices.

        This function:
        - Loads muscle labels and determines the number of distinct muscles.
        - Establishes muscle-to-joint associations for deformation computation.
        - Maps mesh vertices to their corresponding muscles.

        Returns:
            torch.Tensor:
            A tensor representing muscle associations for the horse model.

        """
        # Define the anatomical parts associated with muscles
        muscle_parts = [
            'LScapula',
            'RScapula',
            'Spine1',
            'Spine2',
            'LBLeg1',
            'LBLeg2',
            'LBLeg3',
            'Neck1',
            'Neck2',
            'Neck',
            'Spine',
            'LFLeg1',
            'LFLeg2',
            'LFLeg3',
            'RFLeg2',
            'RFLeg3',
            'RFLeg1',
            'Pelvis',
            'RBLeg2',
            'RBLeg3',
            'RBLeg1',
            'Head'
        ]

        # Aggregate all muscle-related vertices
        all_muscle_idxs = np.concatenate(
            [self.part2vertices[part] for part in muscle_parts]
            )

        self.num_muscles = np.max(self.muscle_labels) + 1
        num_joints = self.get_num_joints()

        # Initialize muscle-to-joint association tensor
        muscle_associations = torch.zeros((num_joints, self.num_muscles))

        # Assign muscle influences to joints based on muscle labels
        for part in muscle_parts:
            # Get joint index for the part
            part_idx = self.parts[part]
            # Retrieve associated vertices
            part_vertices = self.part2vertices[part]
            # Find unique muscle labels
            labels = np.unique(self.muscle_labels[part_vertices])

            # Assign muscles to the corresponding joint
            muscle_associations[part_idx - 1, labels] += 1

            # Also propagate to parent joints
            parent_joint = self.parents[part_idx]
            if parent_joint > -1:
                muscle_associations[parent_joint - 1, labels] += 1

                # Propagate associations to child joints
                for child_idx in np.where(self.parents == part_idx)[0]:
                    muscle_associations[child_idx - 1, labels] += 1

        # Normalize muscle associations for numerical stability
        muscle_associations /= torch.max(muscle_associations)

        # Define the vertex indices associated with each muscle
        muscle_idx_set = set(all_muscle_idxs)
        self.muscle_vertex_map = [
            list(muscle_idx_set & set(np.where(self.muscle_labels == i)[0]))
            for i in range(self.num_muscles)
        ]

        return muscle_associations

    def create_Bm(self):
        """Create Pytorch layers, one for each muscle.

        Each linear layer maps a combination of muscle pose and shape
        parameters into per-vertex deformations for the corresponding muscle.

        Logic:
        - Each muscle gets a corresponding layer, unless the muscle maps to
            zero vertices.
        - Input dimension is based on pose and shape parameters.
        - Output dimension is 3 * number of vertices associated with the
            muscle (x, y, z per vertex).
        - If a muscle does not affect any vertices (out_dim == 0), a
            placeholder `None` is stored at that index in the list instead of
            a real layer (to avoid unnecessary computation).

        Returns:
            torch.nn.ModuleList:
            A list where each entry is either a neural network layer (for
            muscles that have associated vertices) or `None`(for muscles
            that don't).

        Notes:
            - This design choice allows muscles to optionally have no
            influence (e.g., if they correspond to no vertices in the mesh).
            This is safer than adding a "no-op" layer since it avoids
            calling unnecessary zero-output operations.
            - The weight of each linear layer is initialized to a
            normal distribution with mean 0 and standard deviation 0.001.

        """
        num_joints = self.get_num_joints()
        Bm = torch.nn.ModuleList()
        # Append a layer for each muscle
        for i in range(self.num_muscles):
            pose_d = 4
            input_dim = self.muscle_betas_size * num_joints * pose_d + self.shape_betas_for_muscles
            out_dim = len(self.muscle_vertex_map[i]) * 3

            # Avoid No Op layers if the muscle relates to no vertices
            if input_dim > 0 and out_dim > 0:
                layer = nn.Linear(input_dim, out_dim, bias=False)
                torch.nn.init.normal_(layer.weight, mean=0.0, std=0.001)
                Bm.append(nn.Sequential(layer))
            else:
                Bm.append(None)

        return Bm

    @property
    def keypoint_information(self) -> dict[str, int]:
        """Return a dictionary with keypoint names and vertex index.

        This includes keypoints for various parts of the horse's body.

        Returns:
            Dict[str, int]: Dictionary mapping keypoint names to their
            corresponding vertex indices.

        """
        return VERTEX_IDS['varen']


class MuscleBetaPredictor(nn.Module):
    """Predict betas of Muscles based on pose and shape of horse."""

    def __init__(
            self,
            muscle_associations,
            shape_beta_for_muscles,
            dtype=torch.float32
            ):
        """Initialise the Muscle Beta Predictor.

        Args:
            muscle_associations (torch.Tensor):
                Tensor containing muscle associations for the horse model.
            shape_beta_for_muscles (int):
                Number of shape betas for muscle modeling.
            dtype (torch.dtype, optional):
                Data type for the model parameters (default is torch.float32).

        """
        super().__init__()

        self.shape_betas_for_muscles = shape_beta_for_muscles
        self.num_parts, self.num_muscle = muscle_associations.shape
        rot_form_dim = 4  # dimension of rotation -> 4 = quaternion
        self.num_pose = self.num_parts * rot_form_dim

        self.muscledef = nn.Linear(
            self.num_pose + self.shape_betas_for_muscles,
            self.num_muscle,
            bias=False)
        torch.nn.init.normal_(
            self.muscledef.weight,
            mean=0.0,
            std=0.001
            ).to(dtype)
        A_here = torch.zeros(
            self.num_muscle,
            self.num_pose
            ).to(dtype)

        if self.shape_betas_for_muscles > 0:
            A_here = torch.zeros(
                self.num_muscle,
                self.num_pose + self.shape_betas_for_muscles
                )

        for p in range(self.num_parts):
            for k in range(rot_form_dim):
                A_here[:, rot_form_dim * p + k] = muscle_associations[p, :]

        if self.shape_betas_for_muscles > 0:
            A_here[:, self.num_pose:] = 1

        self.A = torch.nn.Parameter(A_here, requires_grad=True).to(dtype)

    def forward(self, pose, betas):
        """Predict the betas and muscle activations for the model.

        Args:
            pose (torch.Tensor):
                The pose parameters for the model.
            betas (torch.Tensor):
                The shape coefficients for the model.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]:
                The predicted muscle betas and the muscle activations.

        """
        pose = pose[:, 3:].view(-1, self.num_parts, 3)
        tensor_b = axis_angle_to_quaternion(pose).view(-1, self.num_parts * 4)
        if self.shape_betas_for_muscles > 0:
            tensor_b = torch.cat(
                (tensor_b, betas[:, :self.shape_betas_for_muscles]),
                dim=1
                )

        tensor_a = self.A * self.muscledef.weight

        tensor_a = tensor_a.unsqueeze(0)
        tensor_a = tensor_a.expand(pose.shape[0], -1, -1)
        tensor_b = tensor_b.unsqueeze(1)
        tensor_b = tensor_b.expand(-1, self.num_muscle, -1)
        betas_muscle = tensor_a * tensor_b

        return betas_muscle, self.A * self.muscledef.weight



"""Linear Blend Skinning (LBS) and related functions."""
# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# You can only use this computer program if you have closed
# a license agreement with MPG or you get the right to use the computer
# program from someone who is authorized to grant you that right.
# Any use of the computer program without a valid license is prohibited and
# liable to prosecution.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# Contact: ps-license@tuebingen.mpg.de

from typing import Optional

import numpy as np
import torch
import torch.nn.functional as F

from .utils import MuscleDeformer, Tensor, rot_mat_to_euler


def find_dynamic_lmk_idx_and_bcoords(  # noqa: PLR0913, PLR0917
    vertices: Tensor,
    pose: Tensor,
    dynamic_lmk_faces_idx: Tensor,
    dynamic_lmk_b_coords: Tensor,
    neck_kin_chain: list[int],
    pose2rot: bool = True,
) -> tuple[Tensor, Tensor]:
    """Compute the faces, barycentric coordinates for the dynamic landmarks.

    To do so, we first compute the rotation of the neck around the y-axis
    and then use a pre-computed look-up table to find the faces and the
    barycentric coordinates that will be used.

    Special thanks to Soubhik Sanyal (soubhik.sanyal@tuebingen.mpg.de)
    for providing the original TensorFlow implementation and for the LUT.

    Parameters
    ----------
    vertices: torch.tensor BxVx3, dtype = torch.float32
        The tensor of input vertices
    pose: torch.tensor Bx(Jx3), dtype = torch.float32
        The current pose of the body model
    dynamic_lmk_faces_idx: torch.tensor L, dtype = torch.long
        The look-up table from neck rotation to faces
    dynamic_lmk_b_coords: torch.tensor Lx3, dtype = torch.float32
        The look-up table from neck rotation to barycentric coordinates
    neck_kin_chain: list
        A python list that contains the indices of the joints that form the
        kinematic chain of the neck.
    pose2rot: bool, optional
        Flag on whether to convert the input pose tensor to rotation

    Returns
    -------
    dyn_lmk_faces_idx: torch.tensor, dtype = torch.long
        A tensor of size BxL that contains the indices of the faces that
        will be used to compute the current dynamic landmarks.
    dyn_lmk_b_coords: torch.tensor, dtype = torch.float32
        A tensor of size BxL that contains the indices of the faces that
        will be used to compute the current dynamic landmarks.

    """
    dtype = vertices.dtype
    batch_size = vertices.shape[0]

    if pose2rot:
        aa_pose = torch.index_select(pose.view(batch_size, -1, 3), 1,
                                     neck_kin_chain)
        rot_mats = batch_rodrigues(
            aa_pose.view(-1, 3)).view(batch_size, -1, 3, 3)
    else:
        rot_mats = torch.index_select(
            pose.view(batch_size, -1, 3, 3), 1, neck_kin_chain)

    rel_rot_mat = torch.eye(
        3, device=vertices.device, dtype=dtype).unsqueeze_(dim=0).repeat(
            batch_size, 1, 1)
    for idx in range(len(neck_kin_chain)):
        rel_rot_mat = torch.bmm(rot_mats[:, idx], rel_rot_mat)

    y_rot_angle = torch.round(
        torch.clamp(-rot_mat_to_euler(rel_rot_mat) * 180.0 / np.pi,
                    max=39)).to(dtype=torch.long)
    neg_mask = y_rot_angle.lt(0).to(dtype=torch.long)
    mask = y_rot_angle.lt(-39).to(dtype=torch.long)
    neg_vals = mask * 78 + (1 - mask) * (39 - y_rot_angle)
    y_rot_angle = (neg_mask * neg_vals +
                   (1 - neg_mask) * y_rot_angle)

    dyn_lmk_faces_idx = torch.index_select(dynamic_lmk_faces_idx,
                                           0, y_rot_angle)
    dyn_lmk_b_coords = torch.index_select(dynamic_lmk_b_coords,
                                          0, y_rot_angle)

    return dyn_lmk_faces_idx, dyn_lmk_b_coords


def vertices2landmarks(
    vertices: Tensor,
    faces: Tensor,
    lmk_faces_idx: Tensor,
    lmk_bary_coords: Tensor
) -> Tensor:
    """Calculate landmarks by barycentric interpolation.

    Parameters
    ----------
    vertices: torch.tensor BxVx3, dtype = torch.float32
        The tensor of input vertices
    faces: torch.tensor Fx3, dtype = torch.long
        The faces of the mesh
    lmk_faces_idx: torch.tensor L, dtype = torch.long
        The tensor with the indices of the faces used to calculate the
        landmarks.
    lmk_bary_coords: torch.tensor Lx3, dtype = torch.float32
        The tensor of barycentric coordinates that are used to interpolate
        the landmarks

    Returns
    -------
    landmarks: torch.tensor BxLx3, dtype = torch.float32
        The coordinates of the landmarks for each mesh in the batch

    """
    # Extract the indices of the vertices for each face
    # BxLx3
    batch_size, num_verts = vertices.shape[:2]
    device = vertices.device
    # The '.to(torch.long)'.
    # added to make the trace work in c++,
    # otherwise you get a runtime error in c++:
    # 'index_select(): Expected dtype int32 or int64 for index'
    lmk_faces = torch.index_select(
        faces,
        0,
        lmk_faces_idx.view(-1).to(torch.long)
        ).view(
        batch_size, -1, 3)

    lmk_faces += torch.arange(
        batch_size, dtype=torch.long, device=device).view(-1, 1, 1) * num_verts

    lmk_vertices = vertices.view(-1, 3)[lmk_faces].view(
        batch_size, -1, 3, 3)

    landmarks = torch.einsum('blfi,blf->bli', [lmk_vertices, lmk_bary_coords])
    return landmarks


def lbs(  # noqa: PLR0913, PLR0914, PLR0917
    betas: Tensor,
    pose: Tensor,
    v_template: Tensor,
    shapedirs: Tensor,
    posedirs: Tensor,
    J_regressor: Tensor,
    parents: Tensor,
    lbs_weights: Tensor,
    pose2rot: bool = True,
    muscle_deformer: Optional[MuscleDeformer] = None,
) -> tuple[Tensor, Tensor]:
    """Perform Linear Blend Skinning with the given shape and pose parameters.

    Parameters
    ----------
    betas: torch.tensor BxNB
        The tensor of shape parameters.
    pose: torch.tensor Bx(J + 1) * 3
        The pose parameters in axis-angle format.
    v_template: torch.tensor BxVx3
        The template mesh that will be deformed.
    shapedirs: torch.tensor Vx3xNB
        The tensor of PCA shape displacements.
    posedirs: torch.tensor Px(V * 3)
        The pose PCA coefficients.
    J_regressor: torch.tensor JxV
        The regressor array that is used to calculate the joints from
        the position of the vertices.
    parents: torch.tensor J
        The array that describes the kinematic tree for the model.
    lbs_weights: torch.tensor Vx(J + 1)
        The linear blend skinning weights that represent how much the
        rotation matrix of each part affects each vertex.
    pose2rot: bool, optional
        Flag on whether to convert the input pose tensor to rotation
        matrices. The default value is True. If False, then the pose tensor
        should already contain rotation matrices and have a size of
        Bx(J + 1)x9.
    muscle_deformer: Optional[MuscleDeformer], optional
        An optional muscle deformer object that applies additional
        deformations to the vertices based on muscle activations.

    Returns
    -------
    verts: torch.tensor BxVx3
        The vertices of the mesh after applying the shape and pose
        displacements.
    joints: torch.tensor BxJx3
        The joints of the model

    """
    batch_size = max(betas.shape[0], pose.shape[0])
    device, dtype = betas.device, betas.dtype

    # 1. Add shape contribution
    v_shaped = v_template + blend_shapes(betas, shapedirs)

    # 2. Get the joints and rotation matrices
    # NxJx3 array
    J = vertices2joints(J_regressor, v_shaped)
    rot_mats = get_rotmats(pose, pose2rot)

    # 3. Get the pose contribution
    if muscle_deformer is not None:
        mdv = neural_pose_deformations(v_shaped, muscle_deformer)
        v_posed = v_shaped + mdv
    else:
        mdv = None
        v_posed = v_shaped + pose_deformations(rot_mats, posedirs)

    # 4. Get the global joint location
    J_transformed, A = batch_rigid_transform(rot_mats, J, parents, dtype=dtype)

    # 5. Do skinning:
    # W is N x V x (J + 1)
    W = lbs_weights.unsqueeze(dim=0).expand([batch_size, -1, -1])
    # (N x V x (J + 1)) x (N x (J + 1) x 16)
    num_joints = J_regressor.shape[0]
    T = torch.matmul(W, A.view(batch_size, num_joints, 16)) \
        .view(batch_size, -1, 4, 4)

    homogen_coord = torch.ones([batch_size, v_posed.shape[1], 1],
                               dtype=dtype, device=device)
    v_posed_homo = torch.cat([v_posed, homogen_coord], dim=2)
    v_homo = torch.matmul(T, torch.unsqueeze(v_posed_homo, dim=-1))

    verts = v_homo[:, :, :3, 0]

    return verts, J_transformed, mdv


def get_rotmats(pose, pose2rot):
    """Convert pose to rotation matrices."""
    batch_size = pose.shape[0]

    if pose2rot:
        return batch_rodrigues(pose.view(-1, 3)).view([batch_size, -1, 3, 3])
    return pose.view(batch_size, -1, 3, 3)


def pose_deformations(rot_mats, posedirs):
    """Convert rotation matrices and weights to 3D deformations."""
    dtype, device = rot_mats.dtype, rot_mats.device
    batch_size = rot_mats.shape[0]

    ident = torch.eye(3, dtype=dtype, device=device)
    pose_feature = (rot_mats[:, 1:, :, :] - ident).view([batch_size, -1])
    pose_offsets = torch.matmul(
        pose_feature, posedirs).view(batch_size, -1, 3)

    return pose_offsets


def neural_pose_deformations(vertices, muscle_deformer):
    """Convert vertices and muscle deformations to 3D deformations."""
    return muscle_deformations(vertices, muscle_deformer)[1]


def vertices2joints(J_regressor: Tensor, vertices: Tensor) -> Tensor:
    """Calculate the 3D joint locations from the vertices.

    Parameters
    ----------
    J_regressor : torch.tensor JxV
        The regressor array that is used to calculate the joints from the
        position of the vertices
    vertices : torch.tensor BxVx3
        The tensor of mesh vertices

    Returns
    -------
    torch.tensor BxJx3
        The location of the joints

    """
    return torch.einsum('bik,ji->bjk', [vertices, J_regressor])


def blend_shapes(betas: Tensor, shape_disps: Tensor) -> Tensor:
    """Calculate the per vertex displacement due to the blend shapes.

    Parameters
    ----------
    betas : torch.tensor Bx(num_betas)
        Blend shape coefficients
    shape_disps: torch.tensor Vx3x(num_betas)
        Blend shapes

    Returns
    -------
    torch.tensor BxVx3
        The per-vertex displacement due to shape deformation

    """
    # Displacement[b, m, k] = sum_{l} betas[b, l] * shape_disps[m, k, l]
    # i.e. Multiply each shape displacement by its corresponding beta and
    # then sum them.
    blend_shape = torch.einsum('bl,mkl->bmk', [betas, shape_disps])
    return blend_shape


def batch_rodrigues(
    rot_vecs: Tensor,
) -> Tensor:
    """Calculate the rotation matrices for a batch of rotation vectors.

    Parameters
    ----------
    rot_vecs: torch.tensor Nx3
        array of N axis-angle vectors

    Returns
    -------
    R: torch.tensor Nx3x3
        The rotation matrices for the given axis-angle parameters

    """
    batch_size = rot_vecs.shape[0]
    device, dtype = rot_vecs.device, rot_vecs.dtype

    angle = torch.norm(rot_vecs + 1e-8, dim=1, keepdim=True)
    rot_dir = rot_vecs / angle

    cos = torch.unsqueeze(torch.cos(angle), dim=1)
    sin = torch.unsqueeze(torch.sin(angle), dim=1)

    # Bx1 arrays
    rx, ry, rz = torch.split(rot_dir, 1, dim=1)
    K = torch.zeros((batch_size, 3, 3), dtype=dtype, device=device)

    zeros = torch.zeros((batch_size, 1), dtype=dtype, device=device)
    K = torch.cat([zeros, -rz, ry, rz, zeros, -rx, -ry, rx, zeros], dim=1) \
        .view((batch_size, 3, 3))

    ident = torch.eye(3, dtype=dtype, device=device).unsqueeze(dim=0)
    rot_mat = ident + sin * K + (1 - cos) * torch.bmm(K, K)
    return rot_mat


def transform_mat(R: Tensor, t: Tensor) -> Tensor:
    """Create a batch of transformation matrices.

    Args:
        R: Bx3x3 array of a batch of rotation matrices
        t: Bx3x1 array of a batch of translation vectors

    Returns:
        T: Bx4x4 Transformation matrix

    """
    # No padding left or right, only add an extra row
    return torch.cat([F.pad(R, [0, 0, 0, 1]),
                      F.pad(t, [0, 0, 0, 1], value=1)], dim=2)


def batch_rigid_transform(
    rot_mats: Tensor,
    joints: Tensor,
    parents: Tensor,
    dtype=torch.float32
) -> Tensor:
    """Apply a batch of rigid transformations to the joints.

    Parameters
    ----------
    rot_mats : torch.tensor BxNx3x3
        Tensor of rotation matrices
    joints : torch.tensor BxNx3
        Locations of joints
    parents : torch.tensor BxN
        The kinematic tree of each object
    dtype : torch.dtype, optional:
        The data type of the created tensors, the default is torch.float32

    Returns
    -------
    posed_joints : torch.tensor BxNx3
        The locations of the joints after applying the pose rotations
    rel_transforms : torch.tensor BxNx4x4
        The relative (with respect to the root joint) rigid transformations
        for all the joints

    """
    joints = torch.unsqueeze(joints, dim=-1)

    rel_joints = joints.clone()
    rel_joints[:, 1:] -= joints[:, parents[1:]]

    transforms_mat = transform_mat(
        rot_mats.reshape(-1, 3, 3),
        rel_joints.reshape(-1, 3, 1)).reshape(-1, joints.shape[1], 4, 4)

    transform_chain = [transforms_mat[:, 0]]
    for i in range(1, parents.shape[0]):
        # Subtract the joint location at the rest pose
        # No need for rotation, since it's identity when at rest
        curr_res = torch.matmul(transform_chain[parents[i]],
                                transforms_mat[:, i])
        transform_chain.append(curr_res)

    transforms = torch.stack(transform_chain, dim=1)

    # The last column of the transformations contains the posed joints
    posed_joints = transforms[:, :, :3, 3]

    joints_homogen = F.pad(joints, [0, 0, 0, 1])

    rel_transforms = transforms - F.pad(
        torch.matmul(transforms, joints_homogen), [3, 0, 0, 0, 0, 0, 0, 0])

    return posed_joints.to(dtype), rel_transforms.to(dtype)


def muscle_deformations(vertices, muscle_deformer):
    """Calculate the muscle deformations for the given vertices."""
    betas_muscle = muscle_deformer.betas_muscle
    Bm = muscle_deformer.Bm
    muscle_idxs = muscle_deformer.muscle_idxs

    mdv = torch.zeros_like(vertices)
    for i in range(len(Bm)):
        if Bm[i] is not None:  # Handle Muscle with no Vertices
            idx = muscle_idxs[i]
            mdv[:, idx] = Bm[i].forward(
                betas_muscle[:, i]
                ).view(
                    vertices.shape[0],
                    -1,
                    3)

    return vertices + mdv, mdv



"""VertexJoinSelector Class."""
# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# You can only use this computer program if you have closed
# a license agreement with MPG or you get the right to use the computer
# program from someone who is authorized to grant you that right.
# Any use of the computer program without a valid license is prohibited and
# liable to prosecution.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# Contact: ps-license@tuebingen.mpg.de


import numpy as np
import torch
from torch import nn

from .utils import to_tensor


# TODO: IMPLEMENT THIS PROPERLY
class VertexJointSelector(nn.Module):
    """Placeholder module for selecting joints from vertices.

    Used currently in HSMAL and SMAL models.

    """

    def __init__(self, **kwargs):
        """Initialise the JointVertexSelector.

        Sets the extra_joints_idxs to an empty array.
        """
        super().__init__()
        _ = kwargs  # to avoid unused argument warning
        extra_joints_idxs = np.array([])

        self.register_buffer('extra_joints_idxs',
                             to_tensor(extra_joints_idxs, dtype=torch.long))

    def forward(self, vertices, joints):
        """Select the joints from the vertices.

        Args:
            vertices: The vertices of the model.
            joints: The joints of the model.

        Returns:
            joints: The selected joints from the vertices.

        NOTE: For VAREN, all joints are currently set under extra_joints..
        TODO: Fix this.

        """
        # The '.to(torch.long)'.
        # added to make the trace work in c++,
        # otherwise you get a runtime error in c++:
        # 'index_select(): Expected dtype int32 or int64 for index'
        extra_joints = torch.index_select(
            vertices,
            1,
            self.extra_joints_idxs.to(torch.long)
            )
        joints = torch.cat([joints, extra_joints], dim=1)

        return joints



import pytest
import torch

from varen.utils import VARENOutput


def test_warning_on_wrong_pose_key(varen_model):
    """Test if a warning is raised when an incorrect pose key is used."""
    with pytest.warns(UserWarning):
        # Using 'pose' instead of 'body_pose'
        varen_model(not_body_pose=torch.zeros((1, 37 * 3)))


def test_no_warning_on_no_input(varen_model):
    """Test if no warning is raised when no input is provided."""
    # This should not raise any warnings
    varen_model()


# Test moving varen model to CPU and CUDA
def test_varen_model_to_cpu1(varen_model):
    """Test if the VAREN model can be moved to CPU."""
    varen_model = varen_model.to('cuda')
    varen_model = varen_model.to('cpu')
    assert varen_model.shapedirs.device == torch.device('cpu'), f"Model should be on CPU, but got {varen_model.shapedirs.device}."
    assert varen_model.betas.device == torch.device('cpu'), f"Model should be on CPU, but got {varen_model.betas.device}."


def test_varen_model_to_cpu2(varen_model):
    """Test if the VAREN model can be moved to CPU."""
    varen_model = varen_model.cuda()
    varen_model = varen_model.cpu()
    assert varen_model.shapedirs.device == torch.device('cpu'), f"Model should be on CPU, but got {varen_model.shapedirs.device}."
    assert varen_model.betas.device == torch.device('cpu'), f"Model should be on CPU, but got {varen_model.betas.device}."


def test_varen_model_to_cuda1(varen_model):
    """Test if the VAREN model can be moved to CUDA."""
    varen_model = varen_model.to('cuda')
    assert varen_model.shapedirs.is_cuda, f"Vertices should be on CUDA, but got {varen_model.shapedirs.device}."
    assert varen_model.betas.is_cuda, f"Vertices should be on CUDA, but got {varen_model.betas.device}."


def test_varen_model_to_cuda2(varen_model):
    """Test if the VAREN model can be moved to CUDA."""
    varen_model = varen_model.to('cuda')
    assert varen_model.shapedirs.is_cuda, f"Vertices should be on CUDA, but got {varen_model.shapedirs.device}."
    assert varen_model.betas.is_cuda, f"Vertices should be on CUDA, but got {varen_model.betas.device}."


# Test CPU inputs on CPU
def test_pose_cpu(varen_model, pose_cpu):
    """Test the model with CPU pose input."""
    varen_model = varen_model.to('cpu')
    output = varen_model(body_pose=pose_cpu)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.device == torch.device('cpu'), f'Vertices should be on CPU, but got {output.vertices.device}.'


def test_global_orient_cpu(varen_model, global_orient_cpu):
    """Test the model with CPU pose input."""
    varen_model = varen_model.to('cpu')
    output = varen_model(global_orient=global_orient_cpu)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.device == torch.device('cpu'), f'Vertices should be on CPU, but got {output.vertices.device}.'


def test_shape_cpu(varen_model, shape_cpu):
    """Test the model with CPU pose input."""
    varen_model = varen_model.to('cpu')
    output = varen_model(betas=shape_cpu)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.device == torch.device('cpu'), f'Vertices should be on CPU, but got {output.vertices.device}.'


def test_transl_cpu(varen_model, translation_cpu):
    """Test the model with CPU pose input."""
    varen_model = varen_model.to('cpu')
    output = varen_model(transl=translation_cpu)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.device == torch.device('cpu'), f'Vertices should be on CPU, but got {output.vertices.device}.'


# Test CUDA inputs on CUDA
def test_pose_cuda(varen_model, pose_cuda):
    """Test the model with CUDA pose input."""
    varen_model = varen_model.to('cuda')
    output = varen_model(body_pose=pose_cuda)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.is_cuda, f"Vertices should be on CUDA, but got {output.vertices.device}."


def test_global_orient_cuda(varen_model, global_orient_cuda):
    """Test the model with CUDA pose input."""
    varen_model = varen_model.to('cuda')
    output = varen_model(global_orient=global_orient_cuda)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.is_cuda, f"Vertices should be on CUDA, but got {output.vertices.device}."


def test_shape_cuda(varen_model, shape_cuda):
    """Test the model with CUDA pose input."""
    varen_model = varen_model.to('cuda')
    output = varen_model(betas=shape_cuda)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.is_cuda, f"Vertices should be on CUDA, but got {output.vertices.device}."


def test_transl_cuda(varen_model, translation_cuda):
    """Test the model with CUDA pose input."""
    varen_model = varen_model.to('cuda')
    output = varen_model(transl=translation_cuda)
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."
    assert output.vertices is not None, "Vertices should not be None."
    assert output.vertices.is_cuda, f"Vertices should be on CUDA, but got {output.vertices.device}."


# Test CUDA inputs on cpu (Should all raise an error)
def test_pose_cuda_on_cpu(varen_model, pose_cuda):
    """Test the model with CUDA pose input on CPU."""
    varen_model = varen_model.to('cpu')
    with pytest.raises(RuntimeError):
        varen_model(body_pose=pose_cuda)


# Test CUDA inputs on cpu (Should all raise an error)
def test_shape_cuda_on_cpu(varen_model, shape_cuda):
    """Test the model with CUDA pose input on CPU."""
    varen_model = varen_model.to('cpu')
    with pytest.raises(RuntimeError):
        varen_model(betas=shape_cuda)


def test_global_orient_cuda_on_cpu(varen_model, global_orient_cuda):
    """Test the model with CUDA global orientation input on CPU."""
    varen_model = varen_model.to('cpu')
    with pytest.raises(RuntimeError):
        varen_model(global_orient=global_orient_cuda)


def test_transl_cuda_on_cpu(varen_model, translation_cuda):
    """Test the model with CUDA translation input on CPU."""
    varen_model = varen_model.to('cpu')
    with pytest.raises(RuntimeError):
        varen_model(transl=translation_cuda)


# Test CPU inputs on CUDA (Should all raise an error)
def test_pose_cpu_on_cuda(varen_model, pose_cpu):
    """Test the model with CPU pose input on CUDA."""
    varen_model = varen_model.to('cuda')
    with pytest.raises(RuntimeError):
        varen_model(body_pose=pose_cpu)


def test_global_orient_cpu_on_cuda(varen_model, global_orient_cpu):
    """Test the model with CPU global orientation input on CUDA."""
    varen_model = varen_model.to('cuda')
    with pytest.raises(RuntimeError):
        varen_model(global_orient=global_orient_cpu)


def test_shape_cpu_on_cuda(varen_model, shape_cpu):
    """Test the model with CPU shape input on CUDA."""
    varen_model = varen_model.to('cuda')
    with pytest.raises(RuntimeError):
        varen_model(betas=shape_cpu)


def test_transl_cpu_on_cuda(varen_model, translation_cpu):
    """Test the model with CPU translation input on CUDA."""
    varen_model = varen_model.to('cuda')
    with pytest.raises(RuntimeError):
        varen_model(transl=translation_cpu)


def test_cuda_input_missing_pose(
        varen_model,
        translation_cuda,
        global_orient_cuda,
        shape_cuda):
    """Test the model with missing pose input on CUDA. This should run."""
    varen_model = varen_model.to('cuda')
    output = varen_model(
        transl=translation_cuda,
        global_orient=global_orient_cuda,
        betas=shape_cuda
    )
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."


def test_cuda_input_missing_global_orient(
        varen_model,
        pose_cuda,
        translation_cuda,
        shape_cuda):
    """Test the model with missing global orientation input on CUDA. This should run."""
    varen_model = varen_model.to('cuda')
    output = varen_model(
        body_pose=pose_cuda,
        transl=translation_cuda,
        betas=shape_cuda
    )
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."


def test_cuda_input_missing_shape(
        varen_model,
        pose_cuda,
        global_orient_cuda,
        translation_cuda):
    """Test the model with missing shape input on CUDA. This should run."""
    varen_model = varen_model.to('cuda')
    output = varen_model(
        body_pose=pose_cuda,
        global_orient=global_orient_cuda,
        transl=translation_cuda
    )
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."


def test_cuda_input_missing_transl(
        varen_model,
        pose_cuda,
        global_orient_cuda,
        shape_cuda):
    """Test the model with missing translation input on CUDA. This should run."""
    varen_model = varen_model.to('cuda')
    output = varen_model(
        body_pose=pose_cuda,
        global_orient=global_orient_cuda,
        betas=shape_cuda
    )
    assert isinstance(output, VARENOutput), "Output should be a ModelOutput instance."


def test_mixed_inputs(
        varen_model,
        pose_cpu,
        global_orient_cuda,
        shape_cuda,
        translation_cuda):
    """Test the model with mixed inputs (CPU pose, CUDA global orientation, CUDA shape, CUDA translation)."""
    varen_model = varen_model.to('cuda')
    with pytest.raises(RuntimeError):
        varen_model(
        body_pose=pose_cpu,
        global_orient=global_orient_cuda,
        betas=shape_cuda,
        transl=translation_cuda
    )
